/**
 * 
 */
package binpackage;

import java.util.List;
import java.util.Vector;
import exceptions.AbortException;
import exceptions.OptimizerException;
import exceptions.NoContextException;

/**
 * Strategy design pattern: This abstract class describes an anonymous optimizer.
 * @author dennis
 */
public abstract class AbstractOptimizer extends ProblemWorker {

	/**
	 * The constructor initializes the generator instance with all data it needs for operation.
	 * All data (including the list of bins) is taken from the given context.
	 * 
	 * @param p_cxt Context object with all necessary data
	 */
	public AbstractOptimizer(Context p_cxt) {
		super(p_cxt);
	}

	/**
	 * This method performs exactly one optimization step. Then it stores its state to the
	 * optimizer object so that it can continue with the next call. It is safe to assume
	 * that the next call to optimizeStep will refer to the same problem. If a new problem
	 * was generated a new optimizer instance would be created.
	 * 
	 * @throws OptimizerException A sub-class of this gets thrown if something goes wrong.
	 * @throws AbortException This exception gets thrown when the optimization reaches its end.
	 *                        This means there's nothing left to optimize.
	 * @throws NoContextException A valid context needs to be provided prior to calling
	 *                            this method.
	 */
	public abstract void optimizerStep() throws OptimizerException, AbortException, NoContextException;
	
	/**
	 * This method is used to query evolutionary optimizers for the amount of dead
	 * offspring born during the last birth phase. Non-evolutionary optimizers don't
	 * need to implement this method since the inherited default behavior is to
	 * return zero. 
	 * @returns The amount of dead offspring during the last optimization step.
	 */
	public int getAmountDeadChildren() {
		return 0;
	}

	List<Context> lEmpty = new Vector<Context>();
	
	/**
	 * This method is used to query evolutionary optimizers for their population.
	 * It's expected that they return the population as a List of Context objects.
	 * Non-evolutionary optimizers don't need to implement this method since the
	 * inherited default behavior is to return an empty list;
	 * @return
	 */
	public List<Context> getPopulation() {
		return this.lEmpty;
	}
	
	/**
	 * This method is used to give an evolutionary optimizer a chance to update all
	 * population member's statistic fields before they are queried for display.
	 * Non-evolutionary optimizers don't need to implement this method since the
	 * inherited default behavior is to just return; 
	 */
	public void updatePopulationStatistics() {
		return;
	}
	
}
