/**
 * 
 */
package binpackage;

import java.util.List;
import java.util.Collections;
import exceptions.AbortException;
import exceptions.GreedyException;
import exceptions.NoContextException;

/**
 * This class implements a greedy solution finder. It works by always assigning the
 * biggest Element to the bin with the most remaining capacity.
 * @author dennis
 *
 */
public final class BigIntoBigGreedy extends AbstractGreedy {

	public static final String strName = "Big to big";
	private double[] dRemainingCapacities;
	private boolean bSorted = false;
	private int iProcessedElements = 0;
	
	/**
	 * Default Constructor.
	 */
	public BigIntoBigGreedy(Context p_cxt) {
		super(p_cxt);
		
		dRemainingCapacities = new double[this.cxt.getAmountBins()];
		this.detectRemainingCapacities(this.cxt.getElements());
		
		this.sortElements(cxt.getElements());
	}

	/**
	 * This method performs exactly one step of the solution finding.
	 * @throws AbortException This is thrown when a solution has been found. It signals
	 *                        that another call to this method will not change the solution.
	 * @throws GreedyException A sub-class of this is thrown when a non-recoverable error
	 *         occurs.
	 * @throws NoContextException This is thrown in case that no valid Context should exist.
	 */
	public void greedyStep() throws GreedyException, AbortException, NoContextException {
		this.checkContext();
		
		try {
			Element eBiggest = findBiggestElement(cxt.getElements());
			int iBiggestBin = this.findBinWithBiggestCapacity();
			
			eBiggest.setBinNr(iBiggestBin);
			this.dRemainingCapacities[iBiggestBin] -= eBiggest.getValue();
			
			// Already signal finish when the last element has been processed.
			iProcessedElements++;
			if (iProcessedElements >= this.cxt.getAmountElements()) {
				throw new NothingFoundException();
			}
			
		} catch (NothingFoundException e) {
			throw new AbortException();
		}

	}
	
	/**
	 * This method sorts all Elements in descending order.
	 */
	private void sortElements(List<Element> p_lElements) {
		Collections.sort(p_lElements);
		Collections.reverse(p_lElements);
		bSorted = true;
	}
	
	/**
	 * This method returns the Element with the biggest value which belongs to no bin
	 * from the given List of Elements.
	 * @param p_lElements
	 * @return
	 */
	private Element findBiggestElement(List<Element> p_lElements) throws NothingFoundException
	{
		boolean bFoundSomething = false;
		Element eBiggest = new Element(0.0);
		
		if (bSorted == true) {
			for (Element e : p_lElements) {
				if (e.getBinNr() < 0) {
					bFoundSomething = true;
					eBiggest = e;
					break;
				}
			}	
		} else {
			for (Element e : p_lElements) {
				if (e.getBinNr() < 0) {
					if (e.getValue() > eBiggest.getValue()) {
						bFoundSomething = true;
						eBiggest = e;
					}
				}
			}
		}
		
		if (bFoundSomething == false) {
			throw new NothingFoundException();
		}
		
		return eBiggest;
	}
	
	private void detectRemainingCapacities(List<Element> p_lElements) {		
		for (int i = 0; i < this.dRemainingCapacities.length; i++) {
			this.dRemainingCapacities[i] = this.cxt.getBinCapacity();
		}
		
		for (Element e : p_lElements) {
			if (e.getBinNr() >= 0) {
				this.dRemainingCapacities[e.getBinNr()] -= e.getValue();
			}
		}
	}
	
	/**
	 * This method goes through the internal array of bin capacities and searches
	 * the entry with the biggest value.
	 * @returns The number of the bin with the biggest remaining capacity.
	 */
	private int findBinWithBiggestCapacity() throws NothingFoundException {
		boolean bFoundSomething = false;
		
		double dCapacity;
		double dBiggestCapacity = Double.MIN_VALUE;
		int iBinNr = 0;
		
		// for (double dCapacity : this.dRemainingCapacities) {
		for (int i = 0; i < this.dRemainingCapacities.length; i++) {
			dCapacity = this.dRemainingCapacities[i];
			
			if (dCapacity > dBiggestCapacity) {
				bFoundSomething = true;
				dBiggestCapacity = dCapacity;
				iBinNr = i;
			}
		}
		
		if (bFoundSomething == false) {
			throw new NothingFoundException();
		}
		
		return iBinNr;
	}
	
	private class NothingFoundException extends Exception {
		static final long serialVersionUID = 1;
	}
}
