/**
 * 
 */
package binpackage;

import java.util.*;

/**
 * Strategy design pattern: Instances of this class hold context specific data
 * needed by the problem generator instances, the greedy solution finder instances
 * and the optimizer instances.
 * @author dennis
 */
public class Context implements Comparable<Context> {

	// Problem specific parameters
	private int iAmountBins = 5;
	private int iAmountElements = 20;
	private double dBinCapacity = 100.0;      // Max. sum of all elements per bin
	private double[] dBinSums;
	
	// Problem representation
	private LinkedList<Element> llElements;
	
	// Optimization specific parameters
	private int iPopulationSize = 10;
	private int iAmountOffspring = 3;
	private int iAmountMutationPerChild = 5;
	
	// Statistics
	private long lLastTimeStamp = 0;
	private long lLastDurationMilliSeconds = 0;
	private long lTotalDurationMilliSeconds = 0;
	private int iAmountDeadChildren = 0;
	private double dMinMissing = 0.0;
	private double dMaxMissing = 0.0;

	/**
	 * Initialize the context with default values
	 */
	public Context() {
		this.initBins(this.iAmountBins, this.iAmountElements);
	}
	
	/**
	 * Initialize context with bin size and amount of elements.
	 * 
	 * @param p_iBins Amount of bins
	 * @param p_iElements Amount of elements
	 */
	public Context(int p_iBins, int p_iElements) {
		this.initBins(p_iBins, p_iElements);
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	protected Object clone() throws CloneNotSupportedException {
		// Make a shallow copy of this object
		Context newCxt = new Context();
		newCxt.iAmountBins = this.iAmountBins;
		newCxt.iAmountElements = this.iAmountElements;
		newCxt.iAmountOffspring = this.iAmountOffspring;
		newCxt.iPopulationSize = this.iPopulationSize;
		newCxt.dBinCapacity = this.dBinCapacity;
		
		newCxt.lLastTimeStamp = this.lLastTimeStamp;
		newCxt.lLastDurationMilliSeconds = this.lLastDurationMilliSeconds;
		newCxt.lTotalDurationMilliSeconds = this.lTotalDurationMilliSeconds;
		newCxt.iAmountDeadChildren = this.iAmountDeadChildren;
		newCxt.dMinMissing = this.dMinMissing;
		newCxt.dMaxMissing = this.dMaxMissing;
		
		// Make a deep copy the contained linked lists
		newCxt.llElements = new LinkedList<Element>();
				
		for (Element e : this.llElements) {
			newCxt.llElements.add(new Element(e));
		}
		
		// Return cloned Context object
		return newCxt;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Fit.: " + this.getFitness();
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Context p_cxt) {
		double dMyFitness = this.getFitness();
		double dHisFitness = p_cxt.getFitness();
		
		if (dMyFitness < dHisFitness) {
			return -1;
		} else if (dMyFitness == dHisFitness) {
			return 0;
		} else if (dMyFitness > dHisFitness) {
			return 1;
		}
		
		return 0;
	}

	/**
	 * This method does a quick and dirty comparison of two Contexts.
	 * @param p_cxt The Context to compare to.
	 * @returns true when the two Contexts might be equal.
	 */
	public boolean quickEquals(Context p_cxt) {
		double dMyFitness;
		double dHisFitness;
		
		// Compare amount of bins and amount of elements
		if (this.getAmountBins() != p_cxt.getAmountBins()) {
			return false;
		}
		
		if (this.getAmountElements() != p_cxt.getAmountElements()) {
			return false;
		}
		
		// Compare fitness
		dMyFitness = this.getFitness();
		dHisFitness = p_cxt.getFitness();
		
		if (dMyFitness != dHisFitness) {
			return false;
		}
		
		// Assume equality
		return true;
	}
	
	/**
	 * This method initializes the bins and the elements. All elements are set to 0.
	 * All bins are empty. They contain no elements.
	 * 
	 * @param p_iBins Amount of bins
	 * @param p_iElements Amount of elements
	 */
	private void initBins(int p_iBins, int p_iElements) {
		// Setting instance variables from method parameters
		this.setAmountBins(p_iBins);
		this.setAmountElements(p_iElements);
		
		// Creating a new linked list for the elements.
		this.llElements = new LinkedList<Element>();
	}
	
	/**
	 * This method calculates the fitness value for the problem which is represented by
	 * this Context objects.
	 * 
	 * It is defined as if:
	 * <pre>
	 * fitness = 0.0;
	 * 
	 * for (bin : bins) {
	 *     fitness += (sumOfAllElementsOfBin - maxCapacityOfBin)^2
	 * }
	 * </pre>
	 * 
	 * @return The fitness value of the problem represented by this Context.
	 */
	public double getFitness() {
		return this.getFitness(this.llElements);
	}
	
	/**
	 * This method calculates the fitness value for the given list of elements.
	 * It assumes that all parameters are identical to those set in the called
	 * Context.
	 * @param p_lElements List of Elements of which to calculate fitness value.
	 * @returns the fitness value for the given List of Elements.
	 */
	public double getFitness(List<Element> p_lElements) {
		double dFitness = 0.0;
		double dMyBinSums[] = new double[this.iAmountBins];
		
		// Make sure the new array start empty
		for (int i = 0; i < dMyBinSums.length; i++) {
			dMyBinSums[i] = 0;
		}
		
		// Calculate sum of all entries per bin
		for (Element e : p_lElements) {
			if (e.getBinNr() >= 0) {
				dMyBinSums[e.getBinNr()] += e.getValue();
			}
		}
		
		// Calculate fitness value
		for (double dBinSum : dMyBinSums) {
			dFitness += Math.pow((dBinSum - this.getBinCapacity()), 2);
		}
		
		return dFitness;	
	}
	
	/**
	 * This method calculates the sum of all Elements of each bins and returns
	 * it as an array of doubles. The list is also stored in the Context for
	 * further retrieval by getBinSums().
	 * @returns The sum of each bin in an double array.
	 */
	public double[] calculateBinSums() {
		this.dBinSums = new double[this.getAmountBins()];
		
		for (int i = 0; i < this.dBinSums.length; i++) {
			this.dBinSums[i] = 0;
		}
		
		for (Element e : this.llElements) {
			if (e.getBinNr() >= 0) {
				this.dBinSums[e.getBinNr()] += e.getValue();
			}
		}
		
		return this.dBinSums;
	}
	
	/**
	 * This method calculates the minimum and the maximum difference of its
	 * bins to their maximum capacity. The result can be read with the methods
	 * getMinMissing() and getMaxMissing().
	 */
	public void calculateMinMaxMissing() {
		double[] dBins = this.calculateBinSums();
		double dMinDifference = this.getBinCapacity();
		double dMaxDifference = 0;
		
		for (double dSum : dBins) {
			double dDifference = Math.abs(this.getBinCapacity() - dSum);
			
			if (dDifference < dMinDifference) {
				dMinDifference = dDifference;
			}
			
			if (dDifference > dMaxDifference) {
				dMaxDifference = dDifference;
			}
		}
		
		this.dMinMissing = dMinDifference;
		this.dMaxMissing = dMaxDifference;
	}
	
	public void startMeasuringDuration() {
		this.lLastTimeStamp = System.currentTimeMillis();
	}
	
	public void stopMeasuringDuration() {
		long lCurrentTimeStamp = System.currentTimeMillis();
		long lDuration = (lCurrentTimeStamp - this.lLastTimeStamp);
		
		this.setLastDuration(lDuration);
		this.lLastTimeStamp = lCurrentTimeStamp;
	}
	
	/**
	 * This method resets the duration of all problem solving steps to zero. 
	 */
	public void resetDuration() {
		this.lLastDurationMilliSeconds = 0;
		this.lTotalDurationMilliSeconds = 0;
	}
	
	/**
	 * This method sets the duration in seconds of the last problem solving
	 * operation. It also adds this duration to the total duration of all
	 * problem solving steps.
	 * @param p_iLastDuration
	 */
	private void setLastDuration(long p_lLastDuration) {
		this.lLastDurationMilliSeconds = p_lLastDuration;
		this.lTotalDurationMilliSeconds += p_lLastDuration;
	}
	
	
	  ///////////////////////////////////////////
	 /// Auto-generated setters and getters ////
	///////////////////////////////////////////
	
	/**
	 * @return the iAmountBins
	 */
	public int getAmountBins() {
		return iAmountBins;
	}

	/**
	 * @param p_iAmountBins the iAmountBins to set
	 */
	public void setAmountBins(int p_iAmountBins) {
		iAmountBins = p_iAmountBins;
	}

	/**
	 * @returns The sum off all Elements per bin in an array of doubles.
	 */
	public double[] getBinSums() {
		if (this.dBinSums == null) {
			this.dBinSums = this.calculateBinSums();
		}
		
		return this.dBinSums;
	}
	
	/**
	 * @return the iAmountElements
	 */
	public int getAmountElements() {
		return iAmountElements;
	}

	/**
	 * @param p_iAmountElements the iAmountElements to set
	 */
	public void setAmountElements(int p_iAmountElements) {
		iAmountElements = p_iAmountElements;
	}
	
	/**
	 * @return the llElements
	 */
	public LinkedList<Element> getElements() {
		return llElements;
	}

	/**
	 * @param p_llElements The list of all Elements to set.
	 */
	public void setElements(LinkedList<Element> p_llElements) {
		this.llElements = p_llElements;
	}
	
	/**
	 * @return the optimizer specific population size.
	 */
	public int getPopulationSize() {
		return iPopulationSize;
	}

	/**
	 * @param p_iPopulationSize Set the optimizer specific population size.
	 */
	public void setPopulationSize(int p_iPopulationSize) {
		iPopulationSize = p_iPopulationSize;
	}

	/**
	 * @return the optimizer specific amount of offspring.
	 */
	public int getAmountOffspring() {
		return iAmountOffspring;
	}

	/**
	 * @param p_iAmountOffspring Sets the optimizer specific amount of offspring.
	 */
	public void setAmountOffspring(int p_iAmountOffspring) {
		iAmountOffspring = p_iAmountOffspring;
	}

	/**
	 * @return the maximum capacity of a bin in units. This means the sum of all
	 * elements of a bin shouldn't be greater than this value.
	 */
	public double getBinCapacity() {
		return this.dBinCapacity;
	}

	/**
	 * @param p_dBinCapacity The maximum capacity of a bin in units. This means the sum of all
	 * elements of a bin shouldn't be greater than this value.
	 */
	public void setBinCapacity(double p_dBinCapacity) {
		this.dBinCapacity = p_dBinCapacity;
	}

	/**
	 * @returns the amount of mutation steps applied in order to bear a child.
	 */
	public int getAmountMutationPerChild() {
		return this.iAmountMutationPerChild;
	}

	/**
	 * @param p_iAmountMutationPerChild The amount of mutation steps applied
	 *                                  in order to bear a child.
	 */
	public void setAmountMutationPerChild(int p_iAmountMutationPerChild) {
		this.iAmountMutationPerChild = p_iAmountMutationPerChild;
	}

	/**
	 * @returns the amount of dead offspring from the last optimization step.
	 */
	public int getAmountDeadChildren() {
		return this.iAmountDeadChildren;
	}

	/**
	 * @param p_iAmountDeadChildren Sets the amount of dead offspring from the
	 *                           last optimization step.
	 */
	public void setAmountDeadChildren(int p_iAmountDeadChildren) {
		this.iAmountDeadChildren = p_iAmountDeadChildren;
	}

	/**
	 * @param p_lDuration Directly sets the duration of the last operation.
	 */
	public void setLastDurationMilliSeconds(long p_lDuration) {
		this.lLastDurationMilliSeconds = p_lDuration;
	}

	/**
	 * @param p_lDuration Directly sets the total duration of all operations.
	 */
	public void setTotalDurationMilliSeconds(long p_lDuration) {
		this.lTotalDurationMilliSeconds = p_lDuration;
	}
	
	/**
	 * @returns the duration in milliseconds of the last operation.
	 */
	public long getLastDurationMilliSeconds() {
		return this.lLastDurationMilliSeconds;
	}

	/**
	 * @returns the total duration in milliseconds of all operations.
	 */
	public long getTotalDurationMilliSeconds() {
		return this.lTotalDurationMilliSeconds;
	}

	/**
	 * @returns the minimum difference of a bin to its full capacity.
	 */
	public double getMinMissing() {
		return this.dMinMissing;
	}

	/**
	 * @return the maximum difference of a bin to its full capacity.
	 */
	public double getMaxMissing() {
		return this.dMaxMissing;
	}

}
