/**
 * 
 */
package binpackage;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.Font;
import java.util.Vector;

/**
 * This class implements a context renderer. This means it takes a Context and outputs
 * a nice visualization of it.
 * @author dennis
 */
public class ContextRenderer {
	
	private Context cxt;
	private double dZoomLevel = 1.0;
	
	private int iRecentWidth = 0;
	private int iRecentHeight = 0;
	
	private int iDefaultBinWidth = 100;
	private int iDefaultSpacing = 10;
	private int iDefaultBinInnerSpacing = 5;
	private int iDefaultFontSize = 14;
	private int iElementDefaultMinWidth = 50;
	private int iElementHeightFactor = 3;

	/**
	 * This class is used internally by the ContextRenderer. It holds information which
	 * are needed to render a certain bin.
	 * @author dennis
	 */
	private class BinRenderingDescription {		
		int iXPos;
		int iYPos;
		int iWidth;
		int iHeight;
		
		int iTopElementYBoundary;
		
		Color color;
		double dSum;
	}
	
	/**
	 * This class is used internally by the ContextRenderer. It holds information which
	 * are needed to render a certain element into a bin.
	 * @author dennis
	 */
	private class ElementRenderingDescription {
		int iXPos;
		int iYPos;
		int iWidth;
		int iHeight;
		
		Color color;
		double dValue;
	}
	
	/*
	 * Default Constructor.
	 */
	public ContextRenderer() {
		this.cxt = null;
	}
	
	/**
	 * This Constructor already initializes the ContextRenderer with a Context.
	 * @param p_cxt The Context to use
	 */
	public ContextRenderer(Context p_cxt) {
		this.setContext(p_cxt);
	}

	/**
	 * This method renders the given Context by creating a new Image object and drawing
	 * to it.
	 * @returns an Image of the rendered Context.
	 */
	public Image render() {
		// Internal variables
		int iAmountBins = this.cxt.getAmountBins();
		
		int iBinWidth = this.iDefaultBinWidth;
		int iSpacing = this.iDefaultSpacing;
		int iBinInnerSpacing = this.iDefaultBinInnerSpacing;
		int iOverHeightOffset = 0;
		
		double dBinCapacity = this.cxt.getBinCapacity();
		int iElementMinWidth = this.iElementDefaultMinWidth;
		int iBinHeight = (int) (iElementHeightFactor * dBinCapacity);
		
		int iFontSize = this.iDefaultFontSize;
		
		Vector<BinRenderingDescription> bins = new Vector<BinRenderingDescription>();
		Vector<ElementRenderingDescription> elements = new Vector<ElementRenderingDescription>();
				
		// Create bin rendering information
		for (int iBin = 0; iBin < iAmountBins; iBin++) {
			BinRenderingDescription bin = new BinRenderingDescription();
			
			bin.iYPos = 3 * iSpacing;
			bin.iXPos = iSpacing + (iBin * (iBinWidth + iSpacing));
			bin.iWidth = iBinWidth;
			bin.iHeight = iBinHeight + 3;
			bin.color = Color.GRAY;
			
			// 3D Border adds one pixel width and height
			bin.iWidth -= 1;
			bin.iHeight -= 1;
			
			bin.dSum = 0.0;
			bin.iTopElementYBoundary = bin.iHeight - 1;
			
			bins.add(bin);
		}
		
		// Create element rendering information
		for (Element e : this.cxt.getElements()) {
			if (e.getBinNr() < 0) {
				continue;
			}
			
			ElementRenderingDescription element = new ElementRenderingDescription();
			BinRenderingDescription bin = bins.get(e.getBinNr());
			
			element.dValue = e.getValue();
			
			element.iWidth = Math.max((bin.iWidth - (2 * iBinInnerSpacing)), iElementMinWidth);
			element.iHeight = (int) (element.dValue * iElementHeightFactor);
			element.iXPos = bin.iXPos + ((bin.iWidth - element.iWidth) / 2);
			element.iYPos = bin.iYPos + bin.iTopElementYBoundary - element.iHeight;
			
			bin.iTopElementYBoundary -= (element.iHeight);
			bin.dSum += element.dValue;
			
			// Detect over-height bins
			if (bin.iTopElementYBoundary < 0) {
				int iSizeDiff = Math.abs(bin.iTopElementYBoundary);
				iOverHeightOffset = Math.max(iOverHeightOffset, iSizeDiff);
			}
			
			// 3D Border adds one pixel width and height
			element.iWidth -= 1;
			element.iHeight -= 1;
			
			int iColor = e.getBinNr() % 5;
			switch(iColor) {
				case 0:  element.color = Color.RED;     break;
				case 1:  element.color = Color.GREEN;   break;
				case 2:  element.color = Color.YELLOW;  break;
				case 3:  element.color = Color.MAGENTA; break;
				case 4:  element.color = Color.ORANGE;  break;
				default: element.color = Color.ORANGE;  break;
			}
			
			elements.add(element);
		}
		
		// Calculate new image size
		int iWidth = iSpacing + (iAmountBins * (iBinWidth + iSpacing));
		int iHeight = iOverHeightOffset + (3 * iSpacing) + iBinHeight + iSpacing;
		
		// Apply zoom level to image size
		iWidth = (int) (iWidth * dZoomLevel);
		iHeight = (int) (iHeight * dZoomLevel);
		iOverHeightOffset = (int) (iOverHeightOffset * dZoomLevel);
		
		// Remember image size and zoom level for fitZoomLevel() method
		this.iRecentWidth = iWidth;
		this.iRecentHeight = iHeight;
		
		// Create a blank image
		BufferedImage img = new BufferedImage(iWidth, iHeight, BufferedImage.TYPE_INT_RGB);
		Graphics2D graph = img.createGraphics();
		graph.setBackground(Color.WHITE);
		graph.clearRect(0, 0, iWidth, iHeight);
		
		// Set font for text display
		iFontSize = (int) (iFontSize * dZoomLevel);
		Font fnt = new Font(Font.SANS_SERIF, Font.BOLD, iFontSize);
		graph.setFont(fnt);
		
		// Render bins
		for (BinRenderingDescription bin : bins) {
			int iDrawXPos = (int) (bin.iXPos * dZoomLevel);
			int iDrawYPos = (int) ((bin.iYPos + iOverHeightOffset) * dZoomLevel);
			int iDrawWidth = (int) (bin.iWidth * dZoomLevel);
			int iDrawHeight = (int) (bin.iHeight * dZoomLevel);
			
			graph.setColor(bin.color);
			graph.fill3DRect(iDrawXPos, iDrawYPos, iDrawWidth, iDrawHeight, true);
		}
		
		
		// Render elements
		for (ElementRenderingDescription element : elements) {
			int iDrawXPos = (int) (element.iXPos * dZoomLevel);
			int iDrawYPos = (int) ((element.iYPos + iOverHeightOffset) * dZoomLevel);
			int iDrawWidth = (int) (element.iWidth * dZoomLevel);
			int iDrawHeight = (int) (element.iHeight * dZoomLevel);
			
			graph.setColor(element.color);
			graph.fill3DRect(iDrawXPos, iDrawYPos, iDrawWidth, iDrawHeight, true);
			
			// Render value into the element
			if (iDrawHeight > (int) (1.5 * iFontSize)) {
				iDrawXPos += (int) (5 * dZoomLevel);
				iDrawYPos += (int) (iDrawHeight - (0.5 * iFontSize));
				
				String strElementValue = "" + (Math.round((element.dValue * 1000000)) / 1000000.0);
				
				graph.setColor(Color.BLACK);
				graph.drawString(strElementValue, iDrawXPos, iDrawYPos);
			}
		}
		
		// Render bin sums above bins
		for (BinRenderingDescription bin : bins) {
			int iDrawXPos = (int) (bin.iXPos * dZoomLevel);
			int iDrawYPos = (int) (2 * iSpacing * dZoomLevel);
			
			String strBinSum = "[" + (Math.round((bin.dSum * 1000000)) / 1000000.0) + "]";
		
			graph.setColor(Color.BLACK);
			graph.drawString(strBinSum, iDrawXPos, iDrawYPos);
		}
		
		// Return image
		return img;
	}

	/**
	 * @return the current Context.
	 */
	public Context getContext() {
		return this.cxt;
	}

	/**
	 * @param cxt The Context to set.
	 */
	public void setContext(Context p_cxt) {
		this.cxt = p_cxt;
	}

	/**
	 * @returns the current zoom level.
	 */
	public double getZoomLevel() {
		return this.dZoomLevel;
	}

	/**
	 * This method sets the zoom level to a given value. 1.0 means no magnification.
	 * @param p_dZoomLevel The new zoom level.
	 */
	public void setZoomLevel(double p_dZoomLevel) {
		if (p_dZoomLevel == 0.0) {
			this.dZoomLevel = 1.0;
		} else {
			this.dZoomLevel = p_dZoomLevel;
		}
	}
		
	/**
	 * This method increases the zoom level by 10%.
	 */
	public void increaseZoomLevel() {
		this.setZoomLevel(this.dZoomLevel * 1.1);
	}
	
	/**
	 * This method decreases the zoom level by 10%.
	 */
	public void decreaseZoomLevel() {
		this.setZoomLevel(this.dZoomLevel / 1.1);
	}

	/**
	 * If an Image has already been rendered this method tries to adjust
	 * the zoom level so that the Image can fit into the given size.
	 * @param p_iWidth Desired maximum width
	 * @param p_iHeight Desired maximum height
	 */
	public void fitZoomLevel(int p_iWidth, int p_iHeight) {
		int iNewWidth;
		int iNewHeight;
		double dScale = 1.0;
		double dScale1 = 1.0;
		double dNewZoomLevel = 1.0;
		
		if ((this.iRecentWidth < 1) || (this.iRecentHeight < 1)) {
			return;
		}
		
		if (p_iWidth < p_iHeight) {
			dScale = (double) p_iWidth / (double) this.iRecentWidth;
			
			iNewWidth = (int) (this.iRecentWidth * dScale);
			iNewHeight = (int) (this.iRecentHeight * dScale);
			
			if (iNewHeight > p_iHeight) {
				dScale1 = (double) p_iHeight / (double) iNewHeight;
				dScale = dScale * dScale1;
				
				iNewWidth = (int) (this.iRecentWidth * dScale1);
				iNewHeight = (int) (this.iRecentHeight * dScale1);
			}
		} else {
			dScale = (double) p_iHeight / (double) this.iRecentHeight;
			
			iNewWidth = (int) (this.iRecentWidth * dScale);
			iNewHeight = (int) (this.iRecentHeight * dScale);
			
			if (iNewWidth > p_iWidth) {
				dScale1 = (double) p_iWidth / (double) iNewWidth;
				dScale = dScale * dScale1;
				
				iNewWidth = (int) (this.iRecentWidth * dScale1);
				iNewHeight = (int) (this.iRecentHeight * dScale1);
			}
		}
		
		dNewZoomLevel = this.getZoomLevel() * dScale;
		this.setZoomLevel(dNewZoomLevel);
	}
}
