/**
 * 
 */
package binpackage;

import java.util.Vector;
import java.util.List;
import java.util.Collections;
import java.util.Random;
import exceptions.AbortException;
import exceptions.NoContextException;
import exceptions.OptimizerException;

/**
 * This class implements a reference implementation of an evolutionary optimizer.
 * @author dennis
 */
public class DarwinOptimizer extends AbstractOptimizer {

	public static final String strName = "Darwin";
	
	private Vector<Context> vCxtPopulation;
	private Vector<Context> vCxtOffspring;
	private Vector<Context> vCxtParents;
	private Random rnd;
	
	private boolean bFirstRun = true;
	private boolean bNoMoreOffspring = false;
	
	private int iAmountDeadChildren = 0;
	
	/**
	 * @param p_cxt
	 */
	public DarwinOptimizer(Context p_cxt) {
		super(p_cxt);
	
		this.vCxtPopulation = new Vector<Context>();
		this.vCxtOffspring = new Vector<Context>();
		this.vCxtParents = new Vector<Context>();
		this.rnd = new Random();
	}

	/* (non-Javadoc)
	 * @see binpackage.AbstractOptimizer#optimizerStep()
	 */
	@Override
	public void optimizerStep() throws OptimizerException, AbortException, NoContextException {
		Context cxtBest;
		
		// Check start conditions
		super.checkContext();
		
		// Late initialization
		if (this.bFirstRun == true) {
			try {
				this.createInitialPopulation(this.cxt);
			} catch (NoMutationPossibleException e) {
				// Already started with an (nearly) optimal solution
				throw new AbortException();
			}
			this.bFirstRun = false;
		}

		// Evolutionary mutation
		this.createOffspring();
		this.deriveNewPopulation();
		
		// Return best solution
		cxtBest = this.returnBestSolution(this.vCxtPopulation);
		
		if (cxtBest != null) {
			this.cxt.setElements(cxtBest.getElements());
		}
		
		// Abort if creation of offspring stopped early
		if (this.bNoMoreOffspring == true) {
			throw new AbortException();
		}
		
		// Abort if optimum found
		if (this.cxt.getFitness() == 0.0) {
			throw new AbortException();
		}
	}
	
	/* (non-Javadoc)
	 * @see binpackage.AbstractOptimizer#getAmountDeadChildren()
	 */
	@Override
	public int getAmountDeadChildren() {
		return this.iAmountDeadChildren;
	}
	
	/* (non-Javadoc)
	 * @see binpackage.AbstractOptimizer#getPopulation()
	 */
	@Override
	public List<Context> getPopulation() {
		return this.vCxtPopulation;
	}

	/* (non-Javadoc)
	 * @see binpackage.AbstractOptimizer#updatePopulationStatistics()
	 */
	@Override
	public void updatePopulationStatistics() {
		long lLastDuration = this.cxt.getLastDurationMilliSeconds();
		long lTotalDuration = this.cxt.getTotalDurationMilliSeconds();
		int iAmountDeadChildren = this.cxt.getAmountDeadChildren();
		
		for (Context c : this.vCxtPopulation) {
			c.setLastDurationMilliSeconds(lLastDuration);
			c.setTotalDurationMilliSeconds(lTotalDuration);
			c.setAmountDeadChildren(iAmountDeadChildren);
			c.calculateMinMaxMissing();
		}
	}

	/**
	 * This method creates an initial population from the given Context.
	 */
	private void createInitialPopulation(Context p_cxt) throws NoMutationPossibleException {
		Context cxtNew = p_cxt;
		boolean bDuplicate;
		int iStart = 0;
		
		this.vCxtPopulation.clear();
		
		
		// Take given solution into population
		try {
			this.vCxtPopulation.add((Context) this.cxt.clone());
			iStart = 1;
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			iStart = 0;
		}
		
		// Derive rest of the population from the given solution
		for (int i = iStart; i < this.cxt.getPopulationSize(); i++) {
			bDuplicate = false;
			
			do {
				cxtNew = this.bearChild(cxtNew);
				bDuplicate = this.checkForContainment(cxtNew, this.vCxtPopulation);
			} while (bDuplicate == true);
			
			this.vCxtPopulation.add(cxtNew);
		}
		
		// Sort population for performance optimization
		this.sortPopulation();
	}
	
	/**
	 * This method uses the current population to create the offspring. No selection
	 * whatsoever takes place at this point. The offspring won't be introduced to the
	 * population.
	 */
	private void createOffspring() {		
		Context cxtChild;
		boolean bDuplicate = false;
		
		// Take the best solutions as parents
		this.vCxtParents.clear();
		
		for (int i = 0; ((i < this.cxt.getPopulationSize()) && (i < this.vCxtPopulation.size())); i++) {
			this.vCxtParents.add(this.vCxtPopulation.get(i));
		}
		
		// Create required offspring from each parent
		this.vCxtOffspring.clear();
		
		// TODO Remove
		System.err.println("Create offspring");
		
		try {
			for (Context cxtParent : vCxtParents) {
				for (int i = 0; i < this.cxt.getAmountOffspring(); i++) {
					do {
						cxtChild = this.bearChild(cxtParent);
						bDuplicate = this.checkForContainment(cxtChild, this.vCxtOffspring);
						
						if (bDuplicate == false) {
							bDuplicate = this.checkForContainment(cxtChild, this.vCxtPopulation);
						}
						
						// TODO Remove
						if (bDuplicate == true)
							System.err.println("Duplicate found. Give birth to another child.");
						
					} while (bDuplicate == true);
						
					this.vCxtOffspring.add(cxtChild);
				}
			}
		} catch (NoMutationPossibleException e) {
			// The population seems to contain a (nearly) optimal solution
			this.bNoMoreOffspring = true;
		}
		
		// Sort offspring for performance optimization
		this.sortOffspring();
	}
	
	/**
	 * This method creates one child from one parent. It does so by first cloning the
	 * parent and than applying a given number of mutations to the child.
	 * @param p_cxtParent The parent Context from which a child shall be derived.
	 * @returns The new born child.
	 * @throws NoMutationPossibleException This Exception signals that no new child
	 *                                     could be created because no improving mutation
	 *                                     could be found.
	 */
	private Context bearChild(Context p_cxtParent)
	throws NoMutationPossibleException {
		Context cxtChild;
		Element element1;
		Element element2;
		
		int iElement1Nr;
		int iElement2Nr;
		int iHamming;
		int iOldBinNr;
		
		boolean bDirection;
		int iStep;
		
		// TODO Remove
		System.err.println("Bear Child from " + p_cxtParent);		
		
		try {
			// Create a new child
			cxtChild = (Context) p_cxtParent.clone();

			// Apply mutation by moving Elements between bins
			for (int i = 0; i < this.cxt.getAmountMutationPerChild(); i++) {
				// Select a random Element
				iElement1Nr = this.rnd.nextInt(cxtChild.getAmountElements());
				element1 = cxtChild.getElements().get(iElement1Nr);
				
				// Find a similar Element with Hamming distance 2
				element2 = element1;
				iElement2Nr = iElement1Nr;
				
				bDirection = this.rnd.nextBoolean();
				
				if (bDirection == true) {
					iStep = 1;
				} else {
					iStep = -1;
				}
				
				iHamming = 0;
				
				while (iHamming < 2) {
					iElement2Nr += iStep;
					
					if ((iElement2Nr < 0) || (iElement2Nr >= cxtChild.getAmountElements())) {
						iStep *= -1;
						iElement2Nr += (2 * iStep);
					}
					
					element2 = cxtChild.getElements().get(iElement2Nr);
					
					if (element1.getBinNr() == element2.getBinNr()) {
						continue;
					}
					
					iHamming = Math.abs(iElement1Nr - iElement2Nr);
				}
				
				// Perform swap
				iOldBinNr = element1.getBinNr();
				element1.setBinNr(element2.getBinNr());
				element2.setBinNr(iOldBinNr);
			}

		} catch (CloneNotSupportedException e) {
			cxtChild = p_cxtParent;
			e.printStackTrace();
		}
		
		// TODO Remove
		System.err.println("New child: " + cxtChild);
		System.err.println();
		
		// Return new child
		cxtChild.calculateBinSums();
		return cxtChild;
	}
	
	/**
	 * This method introduces the offspring to the population. At this point selection
	 * is applied in order to create an improved population. Selection works by picking
	 * the best solutions from the population and the offspring as a new population.
	 */
	private void deriveNewPopulation() {
		Vector<Context> vCxtCandidates = new Vector<Context>();
		
		vCxtCandidates.addAll(this.vCxtPopulation);
		vCxtCandidates.addAll(this.vCxtOffspring);
		Collections.sort(vCxtCandidates);
		
		this.vCxtPopulation.clear();
		for (int i = 0; i < this.cxt.getPopulationSize(); i++) {
			this.vCxtPopulation.add(vCxtCandidates.get(i));
		}
			
		// Sort population for performance optimization
		this.sortPopulation();
		
		// Find out how many children didn't get into the population
		this.iAmountDeadChildren = 0;
		
		for (Context c : this.vCxtOffspring) {
			if (this.vCxtPopulation.contains(c) == false) {
				this.iAmountDeadChildren++;
			}
		}
	}
	
	/**
	 * This method uses the Context.quickEquals() method in order to check whether
	 * a similar Context of the one given is contained in the given List.
	 * 
	 * Attention: This method uses linear searching. So its performance tends to be O(N)!
	 * @param p_cxt The Context to check for containment.
	 * @param p_lCxt The List to check whether a similar Context resides in.
	 * @returns True if a similar Context belongs to the given List.
	 */
	private boolean checkForContainment(Context p_cxt, List<Context> p_lCxt) {
		boolean bDuplicate = false;
		
		for (Context c : p_lCxt) {
			bDuplicate = p_cxt.quickEquals(c);
			
			if (bDuplicate == true) {
				break;
			}
		}
		
		return bDuplicate;
	}
	
	/**
	 * This method scans the given List of Contexts for the best one with the
	 * best fitness value. It assumes that the list is sorted ascending by the
	 * fitness value. This means that the first value will be returned.
	 * @param p_lCxt The list of Contexts in which to search for the best solution.
	 * @returns The Context with the best fitness value.
	 */
	private Context returnBestSolution(List<Context> p_lCxt) {
		if (p_lCxt.size() > 0) {
			return p_lCxt.get(0);
		} else {
			return null;
		}
	}
	
	/**
	 * This method makes sure that the population is sorted ascending by fitness value.
	 */
	private void sortPopulation() {
		Collections.sort(vCxtPopulation);
	}
	
	/**
	 * This method makes sure that the offspring is sorted ascending by fitness value.
	 */
	private void sortOffspring() {
		Collections.sort(vCxtOffspring);
	}
	
	private class NoMutationPossibleException extends Exception {
		public static final long serialVersionUID = 1;
	}
}
