package binpackage;

import java.util.LinkedList;
import java.util.Random;
import exceptions.NoContextException;
import exceptions.ProblemGeneratorException;
import exceptions.ProblemGeneratorLessElementsThanBinsException;

/**
 * This is a simple problem generator. It starts with assuming that each bin is
 * completely filled. Then it removes enough random values from the bins until only
 * one element is missing in each bin and fills the last element with the
 * remaining capacity.
 * It is an almost complete copy of the FastProblemGenerator.
 * @author mfab
 */
public final class FullProblemGenerator extends AbstractProblemGenerator {

	public static final String strName = "Full";
	
	private int i_nBins;                      // Amount of bins
	private int i_nElements;                  // Amount of elements
	
	private Random rnd;                       // Random number generator
	private LinkedList<Element> llElements;   // List of all elements
	
	
	/* (non-Javadoc)
	 * @see binpackage.AbstractProblemGenerator#AbstractProblemGenerator(binpackage.Context)
	 */
	public FullProblemGenerator(Context p_cxt) {
		super(p_cxt);
		
		this.i_nBins = p_cxt.getAmountBins();
		this.i_nElements = p_cxt.getAmountElements();
		
		this.llElements = p_cxt.getElements();
		this.rnd = new Random();
	}
	
	
	/* (non-Javadoc)
	 * @see binpackage.AbstractProblemGenerator#generateProblem()
	 */
	@Override
	public void generateProblem() throws NoContextException,
	                                     ProblemGeneratorException {
		int iBin = -1;
		double newElementValue;
		double[] dBinCapacity = new double[this.i_nBins];
		
		// Check faulty start conditions
		this.checkStartConditions();
		
		if (i_nElements < i_nBins) {
			throw new ProblemGeneratorLessElementsThanBinsException();
		}
		
		// Assume that each bin starts with full remaining capacity
		for (int i = 0; i < dBinCapacity.length; i++) {
			dBinCapacity[i] = this.cxt.getBinCapacity();
		}
		
		/* Progressively remove units from each bins until only one element is 
		 * missing for each bin. It is assured that always less than 75%
		 * of the remaining capacity of a bin are removed from it.
		 */
		for (int i = 0; i < (this.i_nElements-i_nBins); i++) {	
			iBin++;
			iBin = iBin % this.i_nBins;
			
			do {
				newElementValue = (this.rnd.nextDouble() * 0.75) * dBinCapacity[iBin];
			} while (newElementValue == 0.0);
			
			dBinCapacity[iBin] -= newElementValue;
			this.llElements.add(new Element(newElementValue));
		}
		
		// Create last (number of bins) elements by filling each bin to its capacity
		for (int i=0; i < i_nBins; i++) {
			newElementValue = dBinCapacity[i];
			
			dBinCapacity[i] = newElementValue;
			this.llElements.add(new Element(newElementValue));
		}
		
	}

}
