/**
 * 
 */
package binpackage;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Rectangle;
import exceptions.AbortException;
import exceptions.ProblemGeneratorException;
import exceptions.GreedyException;
import exceptions.OptimizerException;

/**
 * This class contains all event handlers for the main frame.
 * @author dennis
 */
public class MainFrameController {
	
	private MainFrame frmMain;
	private ProblemWorkerController pwcController;
	
	/**
	 * Constructor
	 * @param p_frmMain The MainFrame which needs to be controlled.
	 */
	public MainFrameController(MainFrame p_frmMain, ProblemWorkerController p_pwcController) {
		// Selecting frame components
		this.frmMain = p_frmMain;
		this.pwcController = p_pwcController;
						
		// Adding event handlers		
		this.frmMain.btnZoomIn.addActionListener(new ZoomInAction());
		this.frmMain.btnZoomOut.addActionListener(new ZoomOutAction());
		this.frmMain.btnZoomFit.addActionListener(new ZoomFitAction());
		this.frmMain.btnZoomNormal.addActionListener(new ZoomNormalAction());
		
		this.frmMain.btnProblemCreate.addActionListener(new ProblemCreateAction());
		
		this.frmMain.btnGreedySingleStep.addActionListener(new GreedySingleStepAction());
		this.frmMain.btnGreedyStop.addActionListener(new GreedyStopAction());
		this.frmMain.btnGreedyFinish.addActionListener(new GreedyFinishAction());
		
		this.frmMain.btnOptimizerSingleStep.addActionListener(new OptimizerPerformAction());
		this.frmMain.btnOptimizerStop.addActionListener(new OptimizerStopAction());
		this.frmMain.btnOptimizerHold.addActionListener(new OptimizerHoldAction());
		
		this.frmMain.btnDisplaySelectedChild.addActionListener(new DisplayChildAction());
	}
	
	
	/**
	 * This delegate method queries the currently used ProblemWorkerController for
	 * its Context. 
	 * @return the currently employed Context object.
	 */
	protected Context getContext() {
		return this.pwcController.getContext();
	}
	
	/**
	 * This method returns the current problem worker controller.
	 * @returns the current ProblemWorkerController object.
	 */
	protected ProblemWorkerController getProblemWorkerController() {
		return this.pwcController;
	}	
		
	/**
	 * Action listener for changing the displayed population member.
	 * @author dennis
	 */
	protected class DisplayChildAction implements ActionListener {
		public void actionPerformed(ActionEvent arg0) {
			if (frmMain.vCurrentPopulation == null) {
				return;
			}
			
			int ix = frmMain.cbxPopulationBrowser.getSelectedIndex();
			Context cxt = frmMain.vCurrentPopulation.get(ix);
			
			frmMain.updateVisualization(cxt);
		}
	}
	
	
	/**
	 * Action listener for the zoom in tool bar button
	 * @author dennis
	 */
	protected class ZoomInAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			frmMain.crRenderer.increaseZoomLevel();
			frmMain.updateDrawing();
		}
	}

	
	/**
	 * Action listener for the zoom out tool bar button
	 * @author dennis
	 */
	protected class ZoomOutAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			frmMain.crRenderer.decreaseZoomLevel();
			frmMain.updateDrawing();
		}
	}
	
	
	/**
	 * Action listener for the zoom fit tool bar button
	 * @author dennis
	 */
	protected class ZoomFitAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			Rectangle rectVisible = frmMain.lblVisualization.getVisibleRect();
			frmMain.crRenderer.fitZoomLevel(rectVisible.width, rectVisible.height);
			frmMain.updateDrawing();
		}
	}
	
	
	/**
	 * Action listener for the zoom normal tool bar button
	 * @author dennis
	 */
	protected class ZoomNormalAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			frmMain.crRenderer.setZoomLevel(1.0);
			frmMain.updateDrawing();
		}
	}
	
	
	/**
	 * Action listener for the create problem button
	 * @author dennis
	 */
	protected class ProblemCreateAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			// Make sure that no entry field contains invalid data
			frmMain.checkInputFields();
			
			try {
				int iPwIndex = frmMain.cbxProblemStrategy.getSelectedIndex();
				int iAmountBins = (Integer) frmMain.spinProblemAmountBins.getValue();
				int iAmountElements = (Integer) frmMain.spinProblemAmountElements.getValue();
				
				pwcController.createProblem(iPwIndex, iAmountBins, iAmountElements);
			} catch (ProblemGeneratorException e) {
				App.showWarning(e.getMessage());
			}
			
			// Refresh problem visualization
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
		}
	}
		
	
	/**
	 * Action listener for the greedy single step button
	 * @author dennis
	 */
	protected class GreedySingleStepAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			String strMessage = "";
			
			// Make sure that no entry field contains invalid data
			frmMain.checkInputFields();
			
			// Perform one greedy step
			try {
				int iPwIndex = frmMain.cbxGreedyStrategy.getSelectedIndex();
				pwcController.doGreedyStep(iPwIndex);
				
			} catch (GreedyException e) {
				e.printStackTrace();
				App.showWarning(e.getMessage());
			} catch (AbortException e) {
				strMessage = e.getMessage();
			}
			
			// Refresh problem visualization
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
			
			if (! strMessage.equals("")) {
				App.showWarning(strMessage);
			}
		}
	}
	
	
	/**
	 * Action listener for the greedy stop button
	 * @author dennis
	 */
	protected class GreedyStopAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			pwcController.stopGreedy();
			
			// Refresh problem visualization
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
		}
	}
	
	
	/**
	 * Action listener for the finish greedy button
	 * @author dennis
	 */
	protected class GreedyFinishAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			String strMessage = "";
			// Make sure that no entry field contains invalid data
			frmMain.checkInputFields();
			
			// Perform several greedy steps
			int iPwIndex = frmMain.cbxGreedyStrategy.getSelectedIndex();
			
			while (true) {
				try {
					pwcController.doGreedyStep(iPwIndex);
				} catch (AbortException e) {
					strMessage = e.getMessage();
					break;
				} catch (GreedyException e) {
					e.printStackTrace();
					App.showWarning(e.getMessage());
					break;
				}
			}
			
			// Refresh main frame;
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
			
			if (! strMessage.equals("")) {
				App.showWarning(strMessage);
			}
		}
	}
		
	
	/**
	 * Action listener for the optimizer stop button
	 * @author dennis
	 */
	protected class OptimizerStopAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			pwcController.stopOptimizer(true);
			
			// Refresh problem visualization
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
		}
	}
	
	
	/**
	 * Action listener for the optimizer hold button
	 * @author dennis
	 */
	protected class OptimizerHoldAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			pwcController.stopOptimizer(false);
			
			// Refresh problem visualization
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
		}
	}
	
	
	/**
	 * Action listener for the finish optimization button
	 * @author dennis
	 */
	protected class OptimizerPerformAction implements ActionListener {
		public void actionPerformed(ActionEvent p_aeEvent) {
			String strMessage = "";
			
			// Make sure that no entry field contains invalid data
			frmMain.checkInputFields();
			
			// Perform several optimizer steps
			int iPwIndex = frmMain.cbxOptimizerStrategy.getSelectedIndex();
			int iPopulationSize = (Integer) frmMain.spinOptimizerPopulationSize.getValue();
			int iAmountOffspring = (Integer) frmMain.spinOptimizerOffspringSize.getValue();
			int iMutationCount = (Integer) frmMain.spinOptimizerMutationCount.getValue();
			int iSteps = (Integer) frmMain.spinOptimizerSteps.getValue();
			
			for (int i = 0; i < iSteps; i++) {
				try {
					pwcController.doOptimizerStep(iPwIndex, iPopulationSize, iAmountOffspring, iMutationCount);
				} catch (AbortException e) {
					strMessage = e.getMessage();
					break;
				} catch (OptimizerException e) {
					e.printStackTrace();
					App.showWarning(e.getMessage());
					break;
				}
			}
			
			// Refresh main frame;
			frmMain.updateVisualization(getContext());
			frmMain.updatePanelButtons();
			
			if (! strMessage.equals("")) {
				App.showWarning(strMessage);
			}
		}
	}
}
