/**
 * 
 */
package binpackage;

import java.util.LinkedList;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import exceptions.ProblemGeneratorException;
import exceptions.UnableToLoadProblemGeneratorException;
import exceptions.GreedyException;
import exceptions.UnableToLoadGreedyException;
import exceptions.OptimizerException;
import exceptions.UnableToLoadOptimizerException;
import exceptions.NoContextException;
import exceptions.AbortException;

/**
 * This class drives all ProblemWorker instances.
 * @author dennis
 */
public class ProblemWorkerController {

	private LinkedList<ProblemWorkerDescription> llProblemGeneratorDescriptions;
	private LinkedList<ProblemWorkerDescription> llGreedyDescriptions;
	private LinkedList<ProblemWorkerDescription> llOptimizerDescriptions;
	
	private ProblemWorkerState pwsProblemGeneratorState = ProblemWorkerState.STOPPED;
	private ProblemWorkerState pwsGreedyState = ProblemWorkerState.STOPPED;
	private ProblemWorkerState pwsOptimizerState = ProblemWorkerState.STOPPED;
	
	private Context cxt;
	private Context cxtBackup;
	private AbstractGreedy greedyCurrent = null;
	private AbstractOptimizer optimizerCurrent = null;
	
	/**
	 * Default constructor.
	 */
	public ProblemWorkerController() {
		super();
		
		this.llProblemGeneratorDescriptions = new LinkedList<ProblemWorkerDescription>();
		this.llGreedyDescriptions = new LinkedList<ProblemWorkerDescription>();
		this.llOptimizerDescriptions = new LinkedList<ProblemWorkerDescription>();
		
		this.detectProblemWorkers();
	}
	
	public AbstractGreedy getGreedy() {
		return this.greedyCurrent;
	}
	
	public AbstractOptimizer getOptimizer() {
		return this.optimizerCurrent;
	}
	
	/**
	 * This method instantiates a new Context object and generates a new problem.
	 * @param p_iProblemGeneratorIndex Index of the ProblemGenerator to be used.
	 * @param p_iAmountBins Amount of bins.
	 * @param p_iAmountElements Amount of elements.
	 * @throws InfeasibleProblemException The problem could not be created.
	 */
	protected void createProblem(int p_iProblemGeneratorIndex, int p_iAmountBins, int p_iAmountElements) throws ProblemGeneratorException {
		try {
			Context cxtNew = new Context(p_iAmountBins, p_iAmountElements);
			ProblemWorkerDescription pws = this.llProblemGeneratorDescriptions.get(p_iProblemGeneratorIndex);
			AbstractProblemGenerator apg = (AbstractProblemGenerator) pws.getConstructor().newInstance(cxtNew);
			
			// Prepare statistics
			cxtNew.setAmountDeadChildren(0);
			cxtNew.startMeasuringDuration();
			
			// Generate problem
			this.pwsProblemGeneratorState = ProblemWorkerState.RUNNING;
			apg.generateProblem();
			this.pwsProblemGeneratorState = ProblemWorkerState.FINISHED;
			
			// Make sure that a new greedy finder and a new optimizer are used
			this.pwsGreedyState = ProblemWorkerState.STOPPED;
			this.pwsOptimizerState = ProblemWorkerState.STOPPED;
			
			// Finish statistics
			cxtNew.calculateMinMaxMissing();
			cxtNew.stopMeasuringDuration();
			
			this.cxt = cxtNew;
			
		} catch (ProblemGeneratorException e) {
			this.pwsProblemGeneratorState = ProblemWorkerState.STOPPED;
			e.printStackTrace();
			throw e;
		} catch (NoContextException e) {
			e.printStackTrace();
			throw new UnableToLoadProblemGeneratorException();
		} catch (IndexOutOfBoundsException e) {
			e.printStackTrace();
			throw new UnableToLoadProblemGeneratorException();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			throw new UnableToLoadProblemGeneratorException();
		} catch (InstantiationException e) {
			e.printStackTrace();
			throw new UnableToLoadProblemGeneratorException();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			throw new UnableToLoadProblemGeneratorException();
		}
	}
	
	protected void doGreedyStep(int p_iGreedyIndex) throws GreedyException, AbortException {
		GreedyException greedyEx = null;
		AbortException abortEx = null;
		
		try {
			if (this.cxt == null) {
				throw new NoContextException();
			}
			
			if ((this.pwsGreedyState == ProblemWorkerState.STOPPED) || (this.greedyCurrent == null)) {
				// Update Context and store a backup copy
				// No greedy Parameters yet
				this.saveContext();
				
				// Create new greedy solution finder
				ProblemWorkerDescription pws = this.llGreedyDescriptions.get(p_iGreedyIndex);				
				this.greedyCurrent = (AbstractGreedy) pws.getConstructor().newInstance(this.cxt);
				
				// Reset duration measurement
				this.cxt.resetDuration();
			}
			
			this.pwsGreedyState = ProblemWorkerState.RUNNING;
			
			try {
				// Prepare statistics
				this.cxt.setAmountDeadChildren(0);
				this.cxt.startMeasuringDuration();
				
				// Perform greedy step
				this.greedyCurrent.greedyStep();
			} catch (AbortException e) {
				// Set ProblemWorker states
				this.pwsGreedyState = ProblemWorkerState.FINISHED;
				this.pwsOptimizerState = ProblemWorkerState.STOPPED;
				
				abortEx = e;
			}
							
		} catch (GreedyException e) {
			this.stopGreedy();  // Also restores the previous Context
			e.printStackTrace();
			greedyEx = e;
		} catch (NoContextException e) {
			e.printStackTrace();
			greedyEx = new UnableToLoadGreedyException();
		} catch (IndexOutOfBoundsException e) {
			e.printStackTrace();
			greedyEx = new UnableToLoadGreedyException();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			greedyEx = new UnableToLoadGreedyException();
		} catch (InstantiationException e) {
			e.printStackTrace();
			greedyEx = new UnableToLoadGreedyException();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			greedyEx = new UnableToLoadGreedyException();
		}
		
		// Finish statistics
		this.cxt.stopMeasuringDuration();
		this.cxt.calculateMinMaxMissing();
		
		// Throw outstanding exceptions
		if (abortEx != null) {
			throw abortEx;
		}
		
		if (greedyEx != null) {
			throw greedyEx;
		}
	}
	
	protected void doOptimizerStep(int p_iOptimizerIndex, int p_iPopulationSize, int p_iAmountOffspring, int p_iMutationCount) throws OptimizerException, AbortException {
		OptimizerException optEx = null;
		AbortException abortEx = null;
		
		try {
			if (this.cxt == null) {
				throw new NoContextException();
			}
			
			if ((this.pwsOptimizerState == ProblemWorkerState.STOPPED) || (this.optimizerCurrent == null)) {
				// Update Context and store a backup copy
				this.cxt.setPopulationSize(p_iPopulationSize);
				this.cxt.setAmountOffspring(p_iAmountOffspring);
				this.cxt.setAmountMutationPerChild(p_iMutationCount);
				this.saveContext();
				
				// Create new optimizer
				ProblemWorkerDescription pws = this.llOptimizerDescriptions.get(p_iOptimizerIndex);				
				this.optimizerCurrent = (AbstractOptimizer) pws.getConstructor().newInstance(this.cxt);
				
				// Reset duration measurement
				this.cxt.resetDuration();
			}
			
			this.pwsOptimizerState = ProblemWorkerState.RUNNING;
			
			try {
				// Prepare statistics
				this.cxt.setAmountDeadChildren(0);
				this.cxt.startMeasuringDuration();
				
				// Perform optimizer step
				this.optimizerCurrent.optimizerStep();
			} catch (AbortException e) {
				this.pwsOptimizerState = ProblemWorkerState.FINISHED;
				
				/* If there were other ProblemWorkers dependent on the optimizer to be
				 * finished they should be set to STOPPED here.
				 */
				
				abortEx = e;
			}
							
		} catch (OptimizerException e) {
			this.stopOptimizer(true);  // Also restores Context
			e.printStackTrace();
			optEx = e;
		} catch (NoContextException e) {
			e.printStackTrace();
			optEx = new UnableToLoadOptimizerException();
		} catch (IndexOutOfBoundsException e) {
			e.printStackTrace();
			optEx = new UnableToLoadOptimizerException();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			optEx = new UnableToLoadOptimizerException();
		} catch (InstantiationException e) {
			e.printStackTrace();
			optEx = new UnableToLoadOptimizerException();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			optEx = new UnableToLoadOptimizerException();
		}
		
		// Finish statistics
		this.cxt.stopMeasuringDuration();
		this.cxt.calculateMinMaxMissing();
		this.cxt.setAmountDeadChildren(this.optimizerCurrent.getAmountDeadChildren());
		this.optimizerCurrent.updatePopulationStatistics();
		
		// Throw outstanding exceptions
		if (abortEx != null) {
			throw abortEx;
		}
		
		if (optEx != null) {
			throw optEx;
		}
	}
	
	/**
	 * This method returns the currently employed Context.
	 * @return the currently employed Context object.
	 */
	protected Context getContext() {
		return this.cxt;
	}
	
	/**
	 * This method stores a backup copy of the current Context. It is meant to
	 * offer a save state of the problem to return to when the user opts to
	 * stop a solution finding process.
	 */
	protected void saveContext() {
		if (this.cxt != null) {
			try {
				this.cxtBackup = (Context) this.cxt.clone();
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * This method restores the Context from a backup copy. It is meant to
	 * offer a save state of the problem to return to when the user opts to
	 * stop a solution finding process.
	 */
	protected void restoreContext() {
		if (this.cxtBackup != null) {
			this.cxt = this.cxtBackup;
		}
	}
	/**
	 * This method sets the state of the greedy solution finder to stopped. Thus the
	 * solution finder will be marked invalid and be garbage-collected. This method
	 * does not interrupt an already running greedy finder. It just makes sure that it
	 * cannot be re-used.
	 */
	protected void stopGreedy() {
		this.pwsGreedyState = ProblemWorkerState.STOPPED;
		this.restoreContext();
	}
	
	/**
	 * This method sets the state of the optimizer to stopped. Thus the optimizer
	 * will be marked invalid and be garbage-collected. This method does not interrupt
	 * an already running optimizer. It just makes sure that it cannot be re-used.
	 * @param p_bRestore This parameter determines whether the problem will be rolled
	 *                   back to its pre-optimization solution.
	 */
	protected void stopOptimizer(boolean p_bRestore) {
		this.pwsOptimizerState = ProblemWorkerState.STOPPED;
		
		if (p_bRestore == true) {
			this.restoreContext();
		}
	}
	
	/**
	 * This method is used internally to detect all ProblemWorkers and to store their
	 * descriptions.
	 */
	private void detectProblemWorkers() {	
		// Remove all elements from pre-existing lists
		this.llProblemGeneratorDescriptions.clear();
		this.llGreedyDescriptions.clear();
		this.llOptimizerDescriptions.clear();
		
		// Parse all class files
		File fClassDir = new File(Globals.strClassDir);
		File[] fClassFiles = fClassDir.listFiles(new ClassFilenameFilter());
		
		if ((fClassFiles == null) || (fClassFiles.length < 1)) {
			return;
		}
		
		Context cxtDummy = new Context();
		Class<Context> clCxt = Context.class;
		
		for (File fClass : fClassFiles) {
			String strClassName = fClass.getName();
			String strClassPath = Globals.strPackageName + "." + strClassName.subSequence(0, (strClassName.length() - 6));
			
			try {
				Class<?> cl = Class.forName(strClassPath);
				
				// Check for AbstractProblemGenerator
				try {
					Class<AbstractProblemGenerator> apgClass = AbstractProblemGenerator.class;
					cl = cl.asSubclass(apgClass);
					
					Constructor<?> con = cl.getDeclaredConstructor(clCxt);
					AbstractProblemGenerator obj = (AbstractProblemGenerator) con.newInstance(cxtDummy);
					
					Field fldName = cl.getField("strName");
					String strName = (String) fldName.get(obj);
					
					// Add found ProblemGenerator to the list
					ProblemWorkerDescription pwd = new ProblemWorkerDescription(strName, ProblemWorkerType.PROBLEM_GENERATOR, strClassPath, cl, con);
					this.llProblemGeneratorDescriptions.add(pwd);
					
				} catch (ClassCastException e) {
					// Class is no ProblemGenerator
				} catch (NoSuchFieldException e) {
					// Class has no strName field
				} catch (InstantiationException e) {
					// No object can be created
				} catch (IllegalAccessException e) {
					// No object can be created
				} catch (NoSuchMethodException e) {
					// No feasible constructor available
				} catch (InvocationTargetException e) {
					// Unable to create object from class
				}
				
				// Check for AbstractGreedy
				try {
					Class<AbstractGreedy> agClass = AbstractGreedy.class;
					cl = cl.asSubclass(agClass);
					
					Constructor<?> con = cl.getDeclaredConstructor(clCxt);
					AbstractGreedy obj = (AbstractGreedy) con.newInstance(cxtDummy);
					
					Field fldName = cl.getField("strName");
					String strName = (String) fldName.get(obj);
					
					// Add found GreedySolver to the list
					ProblemWorkerDescription pwd = new ProblemWorkerDescription(strName, ProblemWorkerType.GREEDY_SOLVER, strClassPath, cl, con);
					this.llGreedyDescriptions.add(pwd);
					
				} catch (ClassCastException e) {
					// Class is no ProblemGenerator
				} catch (NoSuchFieldException e) {
					// Class has no strName field
				} catch (InstantiationException e) {
					// No object can be created
				} catch (IllegalAccessException e) {
					// No object can be created
				} catch (NoSuchMethodException e) {
					// No feasible constructor available
				} catch (InvocationTargetException e) {
					// Unable to create object from class
				}
				
				// Check for AbstractOptimizer
				try {
					Class<AbstractOptimizer> aoClass = AbstractOptimizer.class;
					cl = cl.asSubclass(aoClass);
					
					Constructor<?> con = cl.getDeclaredConstructor(clCxt);
					AbstractOptimizer obj = (AbstractOptimizer) con.newInstance(cxtDummy);
					
					Field fldName = cl.getField("strName");
					String strName = (String) fldName.get(obj);

					// Add found optimizer to the list
					ProblemWorkerDescription pwd = new ProblemWorkerDescription(strName, ProblemWorkerType.OPTIMIZER, strClassPath, cl, con);
					this.llOptimizerDescriptions.add(pwd);
					
				} catch (ClassCastException e) {
					// Class is no ProblemGenerator
				} catch (NoSuchFieldException e) {
					// Class has no strName field
				} catch (InstantiationException e) {
					// No object can be created
				} catch (IllegalAccessException e) {
					// No object can be created
				} catch (NoSuchMethodException e) {
					// No feasible constructor available
				} catch (InvocationTargetException e) {
					// Unable to create object from class
				}
				
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
				continue;
			}
		}
	}

	  /////////////////////////////////////////////////
	 /// Automatically created Setters and Getters ///
	/////////////////////////////////////////////////
	
	/**
	 * @return a LinkedList of all ProblemWorkerDescriptions of all detected
	 *         problem generators.
	 */
	public LinkedList<ProblemWorkerDescription> getProblemGeneratorDescriptions() {
		return this.llProblemGeneratorDescriptions;
	}

	/**
	 * @return a LinkedList of all ProblemWorkerDescriptions of all detected
	 *         greedy solvers.
	 */
	public LinkedList<ProblemWorkerDescription> getGreedyDescriptions() {
		return this.llGreedyDescriptions;
	}

	/**
	 * @return a LinkedList of all ProblemWorkerDescriptions of all detected
	 *         optimizers.
	 */
	public LinkedList<ProblemWorkerDescription> getOptimizerDescriptions() {
		return this.llOptimizerDescriptions;
	}

	/**
	 * @return the current state of the problem generator.
	 */
	public ProblemWorkerState getProblemGeneratorState() {
		return this.pwsProblemGeneratorState;
	}

	/**
	 * @return the current state of the greedy solution finder.
	 */
	public ProblemWorkerState getGreedyState() {
		return this.pwsGreedyState;
	}
	/**
	 * @return the current state of the optimizer.
	 */
	public ProblemWorkerState getOptimizerState() {
		return this.pwsOptimizerState;
	}
	
}
